/*
 *  Copyright 1995 Science Applications International Corporation.
 */

/*
 * NAME
 *
 * 
 * FILE 
 *	@(#)init_muc.c	6.1
 *
 * SYNOPSIS
 *
 *
 * DESCRIPTION
 *	read the configuration files and the data files.
 *
 * DIAGNOSTICS
 *
 * FILES
 *
 * NOTES
 * 
 * SEE ALSO
 *      parsers.c
 *
 * AUTHOR
 *	Gary Dungca	August 1995
 */

#ifdef SCCSID
static char SccsId[] = "@(#)init_muc.c	6.1\t09/08/97\tSAIC";
#endif

#undef SCCSID
#include  "muc_defs.h"
#include  "muc.h"
#include  "muc_protos.h"
#include  "muc_globals.h"

static void check_config_options(TABLE *option_table);
static void check_map_order(TABLE *option_table);
static void check_scored_slots(TABLE *class_defs);
static void check_map_weights(TABLE *class_defs);
static void check_scoring_task(TABLE *option_table);
static void check_scoring_method(TABLE *option_table);
static void check_single_valued_option(TABLE *option_table, 
				       char *filename_option_name,
				       char *default_filename);
static void check_muc_base_dir(TABLE *option_table);
static void check_file_name(TABLE *option_table,
			    char *filename_option_name,	
			    char *default_filename);
static void check_key_file(TABLE *option_table);
static void check_response_file(TABLE *option_table);
static void check_map_history_file(TABLE *option_table);
static void check_partition_file(TABLE *option_table);
static void check_score_report_file(TABLE *option_table);
static void check_report_summary_file(TABLE *option_table);
static void check_unscored_objects(TABLE *option_table);
static void check_template_name(TABLE *option_table);
static void check_content_name(TABLE *option_table);
static void check_doc_sections(TABLE *option_table);
static void check_dsgs(TABLE *option_table);   /* doc section groups */
static void check_subtask_names(TABLE *option_table);
static void check_equatable_objects(TABLE *option_table);
static void check_postmodifiers(TABLE *option_table);
static void check_premodifiers(TABLE *option_table);
static void check_corporate_designators(TABLE *option_table);
static void check_stringfill_correct_comparison(TABLE *option_table);
static void check_stringfill_partial_comparison(TABLE *option_table);
static void check_optional_status_slot(TABLE *option_table);
static void check_sgml_ID_slot(TABLE *option_table);
static void check_sgml_REF_slot(TABLE *option_table);
static void check_sgml_TEXT_slot(TABLE *option_table);
static void check_sgml_TYPE_slot(TABLE *option_table);
static void check_sgml_alternative_separator(TABLE *option_table);
static void check_report_field_separator(TABLE *option_table);
static void check_sgml_ALT_slot(TABLE *option_table);
static void check_sgml_DOC_gid(TABLE *option_table);
static void check_sgml_DOCNUM_gid(TABLE *option_table);
static void check_sgml_MIN_slot(TABLE *option_table);
static void check_sgml_attribute_quote_char(TABLE *option_table);
static void check_dump_map_history(TABLE *option_table);
static void check_use_IE_report_summary(TABLE *option_table);

static void check_class_defs(TABLE *option_table);
static void check_class_def(char *class_def_string, TABLE *slot_def_strings);
static void parse_class_def_string(char *class_def_string,
				   char **name,
				   char **score_report_name,
				   Bool *scored,
				   int *map_threshold);
static void check_slot_defs(char *classname,
			    TABLE *slot_defs,
			    TABLE *slot_def_strings);
static void check_slot_def(char *classname,
			   TABLE *slot_defs,
			   char *slot_def_string);
static void parse_slot_def_string(char *slot_def_string,
				  char **name,
				  char **score_report_name,
				  Bool *scored,
				  int *map_weight,
				  char **fill_type);




static void initialize_globals(void);
static void open_map_history_file();
static void open_partition_file();
static void open_score_report_file();
static void open_report_summary_file();


void
initialize_MUC_scorer(void)
{
	MUC_Options = table_list_create(option_compare,
					option_release,
					default_destroy);

	parse_config(MUC_configuration_filename);

	check_config_options(MUC_Options);

	initialize_globals();

	parse_keys();
        
	parse_responses();

	open_map_history_file();
    
	if( IS_CO )
		open_partition_file();
	
	open_score_report_file();
	open_report_summary_file();
}    



/************************************************************************
 *
 *  check_config_options
 *	The configuration file has been parsed;  if any required options
 *      were not in the configuration file, either signal an error or 
 *      set the defaults.
 *
 ***********************************************************************/

static
void
check_config_options(TABLE *option_table)
{
	check_scoring_task(option_table);
	check_scoring_method(option_table);

	check_muc_base_dir(option_table);

	check_key_file(option_table);
	check_response_file(option_table);

	check_map_history_file(option_table);
	check_score_report_file(option_table);
	check_report_summary_file(option_table);

	check_class_defs(option_table);
	check_unscored_objects(option_table);
	check_map_order(option_table);
	
	check_template_name(option_table);
	check_content_name(option_table);
	
	if( IS_CO )
	{
		check_partition_file(option_table);
		check_doc_sections(option_table);
	}
	
	if( IS_NE )
	{
		check_doc_sections(option_table);
		check_dsgs(option_table);
		check_subtask_names(option_table);
	}
	
	check_postmodifiers(option_table);
	check_premodifiers(option_table);
	check_corporate_designators(option_table);
	check_stringfill_correct_comparison(option_table);
	check_stringfill_partial_comparison(option_table);
	
	check_optional_status_slot(option_table);
	check_sgml_ID_slot(option_table);
	check_sgml_REF_slot(option_table);
	check_sgml_TEXT_slot(option_table);
	check_sgml_TYPE_slot(option_table);
	check_sgml_alternative_separator(option_table);
	check_report_field_separator(option_table);
	check_sgml_ALT_slot(option_table);
	check_sgml_DOC_gid(option_table);
	check_sgml_DOCNUM_gid(option_table);
	check_sgml_MIN_slot(option_table);
	check_sgml_attribute_quote_char(option_table);
	check_dump_map_history(option_table);
	check_use_IE_report_summary(option_table);
}


static
void
check_map_order(TABLE *option_table)
{
	Option *map_order_option;
	Class_def *class_def;
	
	map_order_option = option_find(option_table, 
				       ":map_order");
	if( map_order_option == NULL )
	{
		map_order_option = option_create(":map_order");

		table_first(MUC_Class_defs);
		for(class_def=(Class_def *)table_next(MUC_Class_defs);
		    class_def;
		    class_def=(Class_def *)table_next(MUC_Class_defs))
			table_append(STRALLOC(class_def->name),
				     map_order_option->values);

		table_append((char *)map_order_option, option_table);
	}
}
	

static
void
check_scoring_task(TABLE *option_table)
{
    char *task_name = option_find_first_value(option_table, ":scoring_task");

    if (task_name ==  NULL)
    {
	fprintf(stderr, "Scoring task undefined.  Cannot continue.\n");
	exit(1);
    }
    
    if (STRCASEEQ(task_name, "scenario_template")
	|| 
	STRCASEEQ(task_name, "scenario-template"))
	    MUC_task = SCENARIO_TEMPLATE;
    else if(STRCASEEQ(task_name, "template_element")
	    ||
	    STRCASEEQ(task_name, "template-element"))
	    MUC_task = TEMPLATE_ELEMENT;
    else if(STRCASEEQ(task_name, "template_relation")
	    ||
	    STRCASEEQ(task_name, "template-relation"))
	    MUC_task = TEMPLATE_RELATION;
    else if(STRCASEEQ(task_name, "named_entity")
	    ||
	    STRCASEEQ(task_name, "named-entity"))
	    MUC_task = NAMED_ENTITY;
    else if(STRCASEEQ(task_name, "coreference"))
	    MUC_task = COREFERENCE;
    else
    {
	fprintf(stderr, "Scoring task '%s' unrecognized.", task_name);
	exit(1);
    }

    return;
}

static
void
check_scoring_method(TABLE *option_table)
{
	Option *method_option;
	TABLE *option_values;
	char *method_name;
	
	if( option_table == NULL )
	{
		fprintf(stderr, "check_scoring_method: null option table \n");
		exit(1);
	}
	
	method_option = option_find(option_table,
				    ":scoring_method");
	if( method_option == NULL )
	{
		method_option = option_create(":scoring_method");
		table_append((char *)method_option, option_table);
	}
	
	option_values = method_option->values;
	if( option_values == NULL )
	{
		fprintf(stderr, 
			"check_scoring_method: option has no value table\n");
		exit(1);
	}
	
	table_first(option_values);
	method_name = table_next(option_values);

	if( method_name == NULL)
	{
		method_name = STRALLOC("key2response");
		table_prepend(method_name, option_values);
	}
	else if(STRCASEEQ(method_name, "key2response")
		||
		STRCASEEQ(method_name, "key2key"))
		;
	else
	{
		fprintf(stderr,
			"check_scoring_method: unknown method: %s\n",
			method_name);
		exit(1);
	}
	if( STRCASEEQ(method_name, "key2response") )
		MUC_method = KEY2RESPONSE;
	else
		MUC_method = KEY2KEY;
}

static
void
check_single_valued_option(TABLE *option_table, 
			   char *option_name,
			   char *default_value)
{
    Option *option;
    TABLE *option_values;
 
    if( option_table == NULL )
    {
	    fprintf(stderr, 
		    "check_single_valued_option: NULL option table\n");
	    exit(1);
    }

    option = option_find(option_table, 
			 option_name);

    if(option == NULL)  /* no option in table */
    {
	    option = option_create(option_name);
	    table_append((char *)option, option_table);
    }
    
    option_values = option->values;

    if(option_values == NULL)  /* shouldn't be possible */
    {
	    fprintf(stderr,
		    "check_single_valued_option: option %s has no values table\n",
		    option_name);
	    exit(1);
    }

    if( table_count(option_values) == 0 )
	    table_append(STRALLOC(default_value), option_values);
}    


static
void
check_muc_base_dir(TABLE *option_table)
{
	check_single_valued_option(option_table,
				   ":muc_base_directory",
				   "");
}

static
void
check_file_name(TABLE *option_table,
		char *option_name,
		char *default_value)
{
    Option *option;
    TABLE *option_values;
    char *muc_base_dir;      /* string to prepend to all files */
    char *given_file_name;   /* what user put in config file */
    char full_file_name[BUFLENGTH]; /* prepended file name */
     
    if( option_table == NULL )
    {
	    fprintf(stderr, 
		    "check_file_name: NULL option table\n");
	    exit(1);
    }

    option = option_find(option_table, 
			 option_name);

    if(option == NULL)  /* no option in table */
    {
	    option = option_create(option_name);
	    table_append((char *)option, option_table);
    }
    
    option_values = option->values;

    if(option_values == NULL)  /* shouldn't be possible */
    {
	    fprintf(stderr,
		    "check_file_name: option %s has no values table\n",
		    option_name);
	    exit(1);
    }

    muc_base_dir = option_find_first_value(option_table,
					   ":muc_base_directory");
    if( muc_base_dir == NULL )
	    full_file_name[0] = '\0';
    else
	    strcpy(full_file_name, muc_base_dir);

    /* get the given file name out of option values table */
    table_first(option_values);  
    given_file_name = table_next(option_values);

    if( given_file_name == NULL )
	    strcat(full_file_name, default_value);
    else
    {
	    strcat(full_file_name, given_file_name);
	    /* delete the un-prepended filename from table */
	    table_delete(default_value, option_values);
    }
    
    /* put full file name into table */
    table_prepend(STRALLOC(full_file_name), option_values);
}    
    
static
void
check_key_file(TABLE *option_table)
{
	check_file_name(option_table,
			":key_file",
			"keys");
}

static
void
check_response_file(TABLE *option_table)
{
	check_file_name(option_table,
			":response_file",
			"responses");
}

static
void
check_map_history_file(TABLE *option_table)
{
	check_file_name(option_table,
			":map_history_file",
			"map_history");
}

static
void
check_partition_file(TABLE *option_table)
{
	check_file_name(option_table,
			":partition_file",
			"partitions");
}


static
void
check_score_report_file(TABLE *option_table)
{
	check_file_name(option_table,
			":score_report_file",
			"scores");
}

static
void
check_report_summary_file(TABLE *option_table)
{
	check_file_name(option_table,
			":report_summary_file",
			"report_summary");
}

static
void
check_optional_status_slot(TABLE *option_table)
{
	check_single_valued_option(option_table,
				   ":optional_status_slot",
				   "status");
}

static
void
check_sgml_ID_slot(TABLE *option_table)
{
	check_single_valued_option(option_table,
				   ":sgml_ID_slot",
				   "id");
}

static
void
check_sgml_REF_slot(TABLE *option_table)
{
	check_single_valued_option(option_table,
				   ":sgml_REF_slot",
				   "REF");
}
	
static
void
check_sgml_TEXT_slot(TABLE *option_table)
{
	check_single_valued_option(option_table,
				   ":sgml_TEXT_slot",
				   "TEXT");
}

static
void
check_sgml_TYPE_slot(TABLE *option_table)
{
	check_single_valued_option(option_table,
				   ":sgml_TYPE_slot",
				   "TYPE");
}

static
void
check_sgml_alternative_separator(TABLE *option_table)
{
	check_single_valued_option(option_table,
				   ":sgml_alternative_separator",
				   "|");
}

static
void
check_report_field_separator(TABLE *option_table)
{
	check_single_valued_option(option_table,
				   ":report_field_separator",
				   "|");
	MUC_report_field_separator 
		= option_find_first_value(MUC_Options,
					  ":report_field_separator");
}

static
void
check_sgml_ALT_slot(TABLE *option_table)
{
	check_single_valued_option(option_table,
				   ":sgml_ALT_slot",
				   "alt");
}

static
void
check_sgml_DOC_gid(TABLE *option_table)
{
	check_single_valued_option(option_table,
				   ":sgml_DOC_gid",
				   "doc");
}

static
void
check_sgml_DOCNUM_gid(TABLE *option_table)
{
	check_single_valued_option(option_table,
				   ":sgml_DOCNUM_gid",
				   "docno");
}

static
void
check_sgml_MIN_slot(TABLE *option_table)
{
	check_single_valued_option(option_table,
				   ":sgml_MIN_slot",
				   "min");
}

static
void
check_sgml_attribute_quote_char(TABLE *option_table)
{
	check_single_valued_option(option_table,
				   ":sgml_attribute_quote_char",
				   "*");
}

static
void
check_dump_map_history(TABLE *option_table)
{
	check_single_valued_option(option_table,
				   ":dump_map_history",
				   "no");
}

static
void
check_use_IE_report_summary(TABLE *option_table)
{
	if( IS_IE )
		check_single_valued_option(option_table,
					   ":use_IE_report_summary",
					   "yes");
	else
		check_single_valued_option(option_table,
					   ":use_IE_report_summary",
					   "no");
}

static
void
check_stringfill_correct_comparison(TABLE *option_table)
{
	char *value;
	check_single_valued_option(option_table,
				   ":stringfill_correct_comparison",
				   "clean");
	value=option_find_first_value(option_table,
				      ":stringfill_correct_comparison");
	if(STRCASEEQ(value,"clean"))
		MUC_stringfill_correct_comparison=CLEAN;
	else if(STRCASEEQ(value,"orig"))
		MUC_stringfill_correct_comparison=ORIG;
	else if(STRCASEEQ(value,"straightened"))
		MUC_stringfill_correct_comparison=STRAIGHTENED;
	else
		MUC_stringfill_correct_comparison=NONE;
}

static
void
check_stringfill_partial_comparison(TABLE *option_table)
{
	char *value;
	check_single_valued_option(option_table,
				   ":stringfill_partial_comparison",
				   "none");
	value=option_find_first_value(option_table,
				      ":stringfill_partial_comparison");
	if(STRCASEEQ(value,"clean"))
		MUC_stringfill_partial_comparison=CLEAN;
	else if(STRCASEEQ(value,"orig"))
		MUC_stringfill_partial_comparison=ORIG;
	else if(STRCASEEQ(value,"straightened"))
		MUC_stringfill_partial_comparison=STRAIGHTENED;
	else
		MUC_stringfill_partial_comparison=NONE;
}

static
void
check_unscored_objects(TABLE *option_table)
{
	Option *unscored_objects_option;
	Class_def *class_def;
	
	unscored_objects_option = option_find(option_table, 
					      ":unscored_objects");
	if( unscored_objects_option == NULL )
	{
		unscored_objects_option = option_create(":unscored_objects");
		table_first(MUC_Class_defs);
		for(class_def=(Class_def *)table_next(MUC_Class_defs);
		    class_def;
		    class_def=(Class_def *)table_next(MUC_Class_defs))
			if( !(class_def->is_scored_class) )
				table_append(STRALLOC(class_def->name),
					     unscored_objects_option->values);
	}
}
	
static
void
check_template_name(TABLE *option_table)
{
	check_single_valued_option(option_table,
				   ":template_name",
				   "template");
}

static
void
check_content_name(TABLE *option_table)
{
	check_single_valued_option(option_table,
				   ":content_name",
				   "content");
}

static
void
check_doc_sections(TABLE *option_table)
{
	Option *doc_sections_option;
	TABLE *doc_sections;
	char *doc_section;
	
	if( option_table == NULL )
	{
		fprintf(stderr, "check_doc_sections: NULL option table\n" );
		exit(1);
	}
	doc_sections_option = option_find(option_table,
					  ":doc_sections");
	if( doc_sections_option == NULL  )
	{
		doc_sections_option = option_create(":doc_sections");
		table_prepend((char *)doc_sections_option, option_table);
	}

	doc_sections = doc_sections_option->values;
	if( doc_sections )
		if(table_count(doc_sections) > 0)
			return;
		else
		{
			table_append(STRALLOC("doc"), 
				     doc_sections);
			table_append(STRALLOC("dateline"),
				     doc_sections);
			table_append(STRALLOC("dd"),
				     doc_sections);
			table_append(STRALLOC("headline"),
				     doc_sections);
			table_append(STRALLOC("text"),
				     doc_sections);
		}
	else
	{
		fprintf(stderr, "check_doc_sections: option has null table\n");
		exit(1);
	}
}

static
void
check_dsgs(TABLE *option_table)
{
/* nothing to check for now */
}

static
void
check_subtask_names(TABLE *option_table)
{
	Option *subtask_names_option;
	TABLE *subtask_names;
	char *subtask_name;
	
	if( option_table == NULL )
	{
		fprintf(stderr, "check_subtask_names: NULL option table\n" );
		exit(1);
	}
	subtask_names_option = option_find(option_table, ":ne_subtask_names");
	if( subtask_names_option == NULL  )
	{
		subtask_names_option = option_create(":ne_subtask_names");
		table_prepend((char *)subtask_names_option, option_table);
	}

	subtask_names = subtask_names_option->values;
	if( subtask_names )
		if(table_count(subtask_names) > 0)
			return;
		else
		{
			table_append(STRALLOC("enamex type organization"),
				     subtask_names);
			table_append(STRALLOC("enamex type person"),
				     subtask_names);
			table_append(STRALLOC("enamex type location"),
				     subtask_names);
			table_append(STRALLOC("enamex type other"),
				     subtask_names);
			table_append(STRALLOC("timex type date"),
				     subtask_names);
			table_append(STRALLOC("timex type time"),
				     subtask_names);
			table_append(STRALLOC("timex type other"),
				     subtask_names);
			table_append(STRALLOC("numex type money"),
				     subtask_names);
			table_append(STRALLOC("numex type percent"),
				     subtask_names);
			table_append(STRALLOC("numex type other"),
				     subtask_names);
		}
	else
	{
		fprintf(stderr, 
			"check_subtask_names: option has null table\n");
		exit(1);
	}
}

static
void
check_equatable_objects(TABLE *option_table )
{
	Option *option;
	
	option = option_find(option_table, ":equatable_objects");
	if( option == NULL )
	{
		option = option_create(":equatable_objects");
		table_append((char *)option, option_table);
	}
}

/*
 * check_corporate_designators -- 
 *        See if the option has been initialized.  Default is none.
 */

static
void
check_corporate_designators(TABLE *option_table)
{
	Option *option;
	TABLE *lowered_values;
	
	option = option_find(option_table, ":corporate_designators");
	if( option == NULL )
	{
		option = option_create(":corporate_designators");
		table_append((char *)option, option_table);
	}
	lowered_values = cstring_lower_table(option->values);
	table_destroy(option->values);
	option->values = lowered_values;
}

static
void
check_postmodifiers(TABLE *option_table)
{
	Option *option = option_find(option_table, ":postmodifiers");
	TABLE *lowered_values;
	
	if( option == NULL )
	{
		option = option_create(":postmodifiers");
		table_append((char *)option, option_table);
	}
	lowered_values = cstring_lower_table(option->values);
	table_destroy(option->values);
	option->values = lowered_values;
}

static
void
check_premodifiers(TABLE *option_table)
{
	Option *option = option_find(option_table, ":premodifiers");
	TABLE *lowered_values;
	
	if( option == NULL )
	{
		option = option_create(":premodifiers");
		table_append((char *)option, option_table);
	}
	lowered_values = cstring_lower_table(option->values);
	table_destroy(option->values);
	option->values = lowered_values;
}

/************************************************************************/

static
void
check_class_defs(TABLE *option_table)
{
	TABLE *class_def_strings;
	char *class_def_string;
	TABLE *slot_def_strings;
	
	class_def_strings = option_find_values(option_table,
					       ":class_defs");
	slot_def_strings = option_find_values(option_table,
					      ":slot_defs");
	
	MUC_Class_defs = table_list_create(class_def_compare,
					   class_def_release,
					   default_destroy);

	if((class_def_strings == NULL) || (slot_def_strings == NULL))
	{
		fprintf(stderr,
			"configuration file missing class or slot definitions\n");
		exit(1);
	}
	
	table_first(class_def_strings);
	for(class_def_string = table_next(class_def_strings);
	    class_def_string;
	    class_def_string = table_next(class_def_strings))
		check_class_def(class_def_string, slot_def_strings);
}


static
void
check_class_def(char *class_def_string, TABLE *slot_def_strings)
{
	Class_def *class_def;
	char *name;
	char *score_report_name;
	Bool scored;
	int map_threshold;
	
	parse_class_def_string(class_def_string,
			       &name,
			       &score_report_name,
			       &scored,
			       &map_threshold);
	
	class_def = class_def_create(name);
	class_def->score_report_name = score_report_name;
	class_def->is_scored_class = scored;
	class_def->map_threshold = map_threshold;
	
	check_slot_defs(name, class_def->slot_defs, slot_def_strings);
	
	table_append((char *)class_def, MUC_Class_defs);
	
	UFREE(name);
}

static
void
parse_class_def_string(char *class_def_string,
		       char **name,
		       char **score_report_name,
		       Bool *scored,
		       int *map_threshold)
{
	char *scored_string;
	char *map_threshold_string;
	char *p1, *p2;
	
	p1 = class_def_string + strspn(class_def_string, " \t");
	p2 = p1 + strcspn(p1, " \t");
	(*name) = alloc_between(p1, p2);
	
	p1 = p2 + strspn(p2, " \t");
	p2 = p1 + strcspn(p1, " \t");
	(*score_report_name) = alloc_between(p1, p2);
	
	p1 = p2 + strspn(p2, " \t");
	p2 = p1 + strcspn(p1, " \t");
	scored_string = alloc_between(p1, p2);
	if( STRCASEEQ(scored_string, "scored"))
		(*scored) = True;
	else
		(*scored) = False;
	
	p1 = p2 + strspn(p2, " \t");
	p2 = p1 + strcspn(p1, " \t");
	map_threshold_string = alloc_between(p1, p2);
	sscanf( map_threshold_string, "%d", map_threshold);
	
	UFREE(scored_string);
	UFREE(map_threshold_string);
}

		
static
void
check_slot_defs(char *classname,
		TABLE *slot_defs,
		TABLE *slot_def_strings)
{
	char *slot_def_string;
	char *p;
	
	table_first(slot_def_strings);
	for(slot_def_string=table_next(slot_def_strings);
	    slot_def_string;
	    slot_def_string=table_next(slot_def_strings))
	{
		p = slot_def_string + strspn(slot_def_string, " \t");
		if( case_is_prefix(p, classname)
		   &&
		   isspace(*(p + strlen(classname)))  )
		{
			p = p + strcspn(p, " \t");
			p = p + strspn(p, " \t");
			check_slot_def(classname, slot_defs, p);
		}
	}
}

static
void
check_slot_def(char *classname,
	       TABLE *slot_defs,
	       char *slot_def_string)
{
	Slot_def *slot_def;
	char *name;
	char *score_report_name;
	Bool scored;
	int map_weight;
	char *fill_type;
	
	parse_slot_def_string(slot_def_string,
			      &name,
			      &score_report_name,
			      &scored,
			      &map_weight,
			      &fill_type);
	
	slot_def = slot_def_create(name, classname);
	slot_def->score_report_name = score_report_name;
	slot_def->is_scored_slot = scored;
	slot_def->map_weight = map_weight;
	table_append((char *)fill_type, slot_def->fill_types);

	table_append((char *)slot_def, slot_defs);
	UFREE(name);
}

	
static
void
parse_slot_def_string(char *slot_def_string,
		      char **name,
		      char **score_report_name,
		      Bool *scored,
		      int *map_weight,
		      char **fill_type)
{
	char *scored_string;
	char *map_weight_string;
	char *p1, *p2;
	
	p1 = slot_def_string + strspn(slot_def_string, " \t");
	p2 = p1 + strcspn(p1, " \t");
	(*name) = alloc_between(p1, p2);
	
	p1 = p2 + strspn(p2, " \t");
	p2 = p1 + strcspn(p1, " \t");
	(*score_report_name) = alloc_between(p1, p2);
	
	p1 = p2 + strspn(p2, " \t");
	p2 = p1 + strcspn(p1, " \t");
	scored_string = alloc_between(p1, p2);
	if( STRCASEEQ(scored_string, "scored"))
		(*scored) = True;
	else
		(*scored) = False;
	
	p1 = p2 + strspn(p2, " \t");
	p2 = p1 + strcspn(p1, " \t");
	map_weight_string = alloc_between(p1, p2);
	sscanf(map_weight_string, "%d", map_weight);
	
	p1 = p2 + strspn(p2, " \t");
	p2 = p1 + strcspn(p1, " \t");
	(*fill_type) = alloc_between(p1, p2);
	
	UFREE(scored_string);
	UFREE(map_weight_string);
}

/************************************************************************/

static
void
initialize_globals(void)
{
	TABLE *ne_subtask_names;
	TABLE *ne_doc_section_names;
	TABLE *dsgs;  /* doc section groups */

	MUC_Catalog = table_list_create(catalog_entry_compare,
					catalog_entry_release,
					default_destroy);
    
	MUC_Summary_Class_scores = table_list_create(class_score_compare,
						     class_score_release,
						     default_destroy);
	class_score_init_table(MUC_Summary_Class_scores, "");

	MUC_Total_Slot_tallies = tallies_create();

	MUC_Total_Text_Filtering = tallies_create();

	ne_subtask_names = option_find_values(MUC_Options, ":ne_subtask_names");
	MUC_ne_subtask_tallies = named_tallies_table_create(ne_subtask_names);
							      
	ne_doc_section_names = option_find_values(MUC_Options, 
						  ":doc_sections");
	MUC_ne_doc_section_tallies = 
		named_tallies_table_create(ne_doc_section_names);

	dsgs = option_find_values(MUC_Options,
				  ":doc_section_groups");
	if( dsgs )
		MUC_dsg_tallies = 
			named_tallies_table_create(dsgs);
	else
		MUC_dsg_tallies = NULL;

	MUC_total_key_eqclasses = 0;
	MUC_total_response_eqclasses = 0;
}

static
void
open_map_history_file()
{
	char *dump_map_history;

	dump_map_history = option_find_first_value(MUC_Options,
						   ":dump_map_history");

	if( STRCASEEQ(dump_map_history, "no") )
		return;
	
	MUC_map_history_filename = 
		option_find_first_value(MUC_Options,
					":map_history_file");
	MUC_map_history_file = 
		fopen(MUC_map_history_filename, "w");
	
	if(MUC_map_history_file == NULL)
	{
		fprintf(stderr, "Could not open map history file '%s'.\n",
			MUC_map_history_filename);
		exit(1);
	}
}

static
void
open_partition_file()
{
	MUC_partition_filename = 
		option_find_first_value(MUC_Options,
					":partition_file");
	MUC_partition_file = 
		fopen(MUC_partition_filename, "w");
	
	if(MUC_partition_file == NULL)
	{
		fprintf(stderr, "Could not open partition file '%s'.\n",
			MUC_partition_filename);
		exit(1);
	}
}

static
void
open_score_report_file()
{
	MUC_score_report_filename = 
		option_find_first_value(MUC_Options,
					":score_report_file");
	MUC_score_report_file = 
		fopen(MUC_score_report_filename, "w");
	
	if(MUC_score_report_file == NULL)
	{
		fprintf(stderr, "Could not open score report file '%s'.\n",
			MUC_score_report_filename);
		exit(1);
	}
}


static
void
open_report_summary_file()
{
	MUC_report_summary_filename = 
		option_find_first_value(MUC_Options,
					":report_summary_file");
	MUC_report_summary_file = 
		fopen(MUC_report_summary_filename, "w");
	
	if(MUC_report_summary_file == NULL)
	{
		fprintf(stderr, "Could not open report summary file '%s'.\n",
			MUC_score_report_filename);
		exit(1);
	}
}
