/*
 *  Copyright 1995 Science Applications International Corporation.
 */

/*
 * NAME
 * FILE 
 *	inst_pair.c
 *
 * SYNOPSIS
 * DESCRIPTION
 *	support routines for Inst_pair data structure.
 *
 * DIAGNOSTICS
 * FILES
 * NOTES
 *
 * SEE ALSO
 * AUTHOR
 *	Gary Dungca	August 1995
 *
 */


#ifdef SCCSID
static char SccsId[] = "@(#)inst_pair.c	6.1\t09/08/97\tSAIC";
#endif

#undef SCCSID
#include  "muc_defs.h"
#include  "muc.h"
#include  "muc_protos.h"
#include  "muc_globals.h"


/************************************************************************
 *									*
 *  inst_pair_create							*
 *									*
 ***********************************************************************/

Inst_pair *
inst_pair_create(Inst *key_inst, Inst *response_inst, char status)
{
    Inst_pair *inst_pair;
    
    if((inst_pair = UALLOC(Inst_pair, 1)) == (Inst_pair *) NULL)
    {
	fprintf(stderr, "Error: allocation for inst_pair failed.\n");
	exit(1);
    }

    inst_pair->key_inst      = key_inst;
    inst_pair->response_inst = response_inst;
    inst_pair->status        = status;
    inst_pair->slot_pairs    = table_list_create(default_compar,
						 slot_pair_release,
						 default_destroy );
    inst_pair->tallies =  tallies_create();
    inst_pair->score =    0.0 ;
    
    return(inst_pair);
}


/************************************************************************
 *									*
 *  inst_pair_release						*
 *									*
 ***********************************************************************/

void
inst_pair_release(char *elt)
{
    Inst_pair *ptr = (Inst_pair *) elt;

    UFREE(ptr->tallies);
    table_destroy( ptr->slot_pairs );
    UFREE(ptr);
}


/************************************************************************
 *									*
 *  inst_pair_sum_tallies
 *									*
 ***********************************************************************/

void
inst_pair_sum_tallies(Tallies *tallies, TABLE *inst_pairs )
{
  Inst_pair *inst_pair;
  
  table_first(inst_pairs);
  for(inst_pair=(Inst_pair *)table_next(inst_pairs);
      inst_pair;
      inst_pair=(Inst_pair *)table_next(inst_pairs))
    tallies_combine( tallies, inst_pair->tallies );
}



/************************************************************************
 *									*
 *  inst_pair_f_compare
 *									*
 *  used to compare inst_pair structures during "qsort"-ing.
 *       of "correct" insts.
 *									*
 ***********************************************************************/

int
inst_pair_f_compare(const void *elt1, const void *elt2)
{
    Inst_pair *ptr1 = *(Inst_pair **)elt1;
    Inst_pair *ptr2 = *(Inst_pair **)elt2;

    int dif;
    
    /* qsort will normally sort in ascending order.  however,
       higher f-scores indicate better mapping results, so an
       a descending order is preferred.  reversing the order
       of comparison is one way, but i will instead apply the
       unary minus to the result of normal comparison. */

    dif = (int) -(ptr1->score - ptr2->score);

    return(dif);
}


/************************************************************************
 *									*
 *  inst_pair_wrong_compare
 *									*
 *  used to compare inst_pair structures during "qsort"-ing of
 *       "incorrect" insts -- we want incorrect inst pairs sorted
 *       in increasing order by the number of wrong tallies
 *       (incorrect + missing + spurious).
 *									*
 ***********************************************************************/

int
inst_pair_wrong_compare(const void *elt1, const void *elt2)
{
    Inst_pair *ptr1 = *(Inst_pair **)elt1;
    Inst_pair *ptr2 = *(Inst_pair **)elt2;

    int dif;
    
    dif = (int) (  ptr1->tallies->incorrect 
		 + ptr1->tallies->missing
		 + ptr1->tallies->spurious
		 - ptr2->tallies->incorrect 
		 - ptr2->tallies->missing
		 - ptr2->tallies->spurious);

    return(dif);
}

/************************************************************************
 *									*
 *  inst_pair_offset_compare
 *									*
 *  used to compare inst_pair structures during "qsort"-ing.
 *     of chosen NE insts (see align_NE.c)
 *									*
 ***********************************************************************/

int
inst_pair_offset_compare(const void *elt1, const void *elt2)
{
    Inst_pair *ip1 = *(Inst_pair **)elt1;
    Inst_pair *ip2 = *(Inst_pair **)elt2;
    Inst *inst1;
    Inst *inst2;

    if( ip1 )
    {
	    if( ip1->key_inst )
		    inst1 = ip1->key_inst;
	    else 
		    inst1 = ip1->response_inst;
    }
    
    if( ip2 )
    {
	    if( ip2->key_inst )
		    inst2 = ip2->key_inst;
	    else
		    inst2 = ip2->response_inst;
    }
    
    if( (inst1==NULL) && inst2 ) return -1;
    if( inst1 && (inst2==NULL) ) return 1;
    if( (inst1==NULL) && (inst2==NULL) ) return 0;
    
    if( inst1->start_offset < inst2->start_offset ) return -1;
    if( inst2->start_offset < inst1->start_offset ) return 1;
    
    return( inst2->end_offset - inst2->end_offset );
}

void
inst_pair_table_2_array(Inst_pair **array, TABLE *inst_pairs)
{
	int i;
	Inst_pair *inst_pair;
	
	table_first(inst_pairs);
	for(i=0, inst_pair=(Inst_pair *)table_next(inst_pairs);
	    inst_pair;
	    i++, inst_pair=(Inst_pair *)table_next(inst_pairs))
		array[i] = inst_pair;
}		

void 
inst_pair_array_2_table(TABLE *inst_pairs, Inst_pair **array, int n)
{
	int i;
	
	for(i=0; i<n; i++)
		table_append((char *)array[i], inst_pairs);
}


/************************************************************************
 *									*
 *  inst_pair_print_table
 *
 ***********************************************************************/

void
inst_pair_print_table(TABLE *inst_pairs, int indent)
{
	Inst_pair *inst_pair;
  
	printf("%*.sinst_pairs:\n", indent, "");
	table_first(inst_pairs);
	for(inst_pair = (Inst_pair *)table_next(inst_pairs);
	    inst_pair;
	    inst_pair = (Inst_pair *)table_next(inst_pairs))
		inst_pair_print(inst_pair, indent+2);
}

/************************************************************************
 *									*
 *  inst_pair_print							*
 *									*
 ***********************************************************************/

void
inst_pair_print(Inst_pair *inst_pair, int indent)
{
	if( inst_pair->key_inst )
		printf("<%s-%s-%s>", 
		       inst_pair->key_inst->name,
		       inst_pair->key_inst->docnum,
		       inst_pair->key_inst->id);
	tallies_print(inst_pair->tallies);
	if( inst_pair->response_inst )
		printf("<%s-%s-%s>\n", 
		       inst_pair->response_inst->name,
		       inst_pair->response_inst->docnum,
		       inst_pair->response_inst->id);
}

