/*
 *  Copyright 1995 Science Applications International Corporation.
 */

/*
 * NAME
 *
 * 
 * FILE 
 *	muc_strings.c
 *
 * SYNOPSIS
 *
 *
 * DESCRIPTION
 *	string manipulation routines.
 *
 * DIAGNOSTICS
 *
 *
 * FILES
 *
 *
 * NOTES
 * 
 * SEE ALSO
 *
 * AUTHOR
 *	Gary Dungca	August 1995
 *
 */


#ifdef SCCSID
static char SccsId[] = "@(#)muc_strings.c	6.1\t09/08/97\tSAIC";
#endif

#undef SCCSID
#include  "muc_defs.h"
#include  "muc.h"
#include  "muc_protos.h"
#include  "muc_globals.h"

static int muc_min_streq(char *kstr, char *rstr, char *mstr);
static char *get_minstring(char *src);
static Bool is_postmodifier(char *string);


Bool
case_is_prefix(char *s, char *c)
{
	if( s == NULL ) return False;
	if( c == NULL ) return False;

	while( *s && *c && case_eq(*s, *c) )
	{
		s++;
		c++;
	}
	if( *c == '\0' ) 
		return True;
	else
		return False;
}

Bool
case_is_suffix(char *whole, char *string)
{
	char *w, *s;
	
	w = whole  + strlen(whole);
	s = string + strlen(string);
	
	while( case_eq(*w,*s) && (w > whole) && (s > string))
	{
		w--;
		s--;
	}
	if( (s == string) && (*w == *s) )
		return True;
	else
		return False;
}

/***********************************************************************
 *
 *  alloc_between -- 
 *         allocate memory and copy the string between firstchar
 *         and the char *before* afterlast inclusive
 *         example:
 *       
 *         123456789012345
 *            ^    ^
 *    firstchar    afterlast
 *
 *
 *            45678
 *            ^
 *       result
 *
 ***********************************************************************
 */

char *
alloc_between( char *firstchar, char *afterlast )
{
	char *result;
	
	if(firstchar >= afterlast)
	{
		result = (char *) malloc(1);
		*result = '\0';
	}
	else
	{
		result = (char *) malloc(afterlast - firstchar + 1);
		if( result == NULL )
		{
			fprintf(stderr,
				"[alloc_between]:can't malloc!\n");
			exit(1);
		}
		memcpy( result, firstchar, afterlast-firstchar );
		result[afterlast-firstchar] = '\0';
	}
	return result;
}

/************************************************************************
 *					
 *  trim_quotes				
 *					
 *  take a quoted string and return a copy of the string, which has had
    any whitespace removed from the beginning and end, then any matching
    quotes removed from the beginning and end, and then any whitespace
    removed from the beginning and end.

    example (colons are to make leading whitespace "visible"):

    given:      :   "  John "JD" Dooner   "   :
    return:     :John "JD" Dooner:
 * 
 * 
 ***********************************************************************/

char *
trim_quotes(char *quoted_string)
{
	char *first_cut;        /* ws trimmed before and after quotes */
	char *second_cut;       /* quotes trimmed */
	char *result;           /* ws trimmed inside quotes */
	char *first_char;       /* pointers to first and char after last */
	char *after_last;
	
	first_cut = trim_whitespace(quoted_string);
	
	first_char = first_cut;
	after_last = first_cut + strlen(first_cut); /* points to '\0' */
	
	if((*first_char == *(after_last - 1))      /* chars of equal value */
	   &&
	   (first_char != (after_last-1))          /* pointers not the same*/
	   &&
	   ((*first_char == '\"')                  /* " chars are quotes */
	    ||
	    (*first_char == '\'')))
	{
		first_char++;
		after_last--;
	}
	
	second_cut = alloc_between(first_char, after_last);
	result = trim_whitespace(second_cut);
	UFREE(first_cut);
	UFREE(second_cut);
	return result;
}


/************************************************************************
 *									*
 *  trim_whitespace							*
 *									*
 *  returns a copy of str that has no leading or trailing whitespace    *
 *									*
 ***********************************************************************/

char *
trim_whitespace(char *str)
{
	char *firstchar, *afterlast;
	
	firstchar = str;
	firstchar += strspn(firstchar, " \t");  /* skip leading whitespace */

	afterlast = str + strlen(str);
	while((*(afterlast-1) == ' ')           /* skip trailing whitespace */
	      ||
	      (*(afterlast-1) == '\t'))
		afterlast--;
	return alloc_between(firstchar, afterlast);
}

/************************************************************************
 *									*
 *  trim_intertoken_whitespace
 *									*
 *  returns a copy of str that has no leading or trailing whitespace    *
 *  and has only a single space between words. 
 *									*
 ***********************************************************************/

char *
trim_intertoken_whitespace(char *str)
{
    char *tokenstr, *result, *token;
    
    tokenstr = STRALLOC(str);                  /* copy to munch on   */
    result = (char *)malloc(strlen(str)+1);    /* alloc result space */
    result[0] = '\0';                          /* prepare for strcat */

    /* grab the first token (no whitespace before it) */
    if(token = strtok(tokenstr, " \t"))
    {
	strcat(result, token);
	/* extract tokens until there are no more (add whitespace before) */
	while(token = strtok((char *)NULL, " \t"))
	{
		strcat(result, " ");
		strcat(result, token);
	}
    }

    UFREE(tokenstr);
    return result;
}


/************************************************************************
 *									*
 *  string_lower							*
 *									*
 *  allocates and returns a lower-case version of STR.			*
 *  this is primarily for use in string routines that need		*
 *  case-independence.                                                  *
 *									*
 ***********************************************************************/

char *
string_lower(char *str)
{
    char *result, *resultptr;
    int length;

    result = STRALLOC(str);
    resultptr = result;
    
    while(*resultptr != '\0')
    { 
	    /* explicitly test for uppercase so we can be "8-bit clean" */
	    if((*resultptr >= 'A') && (*resultptr <= 'Z'))
		    *resultptr = (char)muc_tolower(*resultptr);
	    resultptr++;
    }
    return(result);
}

/************************************************************************
 *									*
 *  nice_docnum								*
 *									*
 *  take an extracted docnum string (from an SGML document, say), and	*
 *  return a new string that contains only alphanumeric chars    	*
 *									*
 ***********************************************************************/

char *
nice_docnum(char *docnum)
{
    char *result;
    char *result_ptr;
    char *docnum_ptr;
    
    if ((result = (char *) malloc (strlen(docnum) +1)) == (char *)NULL)
    {
	fprintf(stderr, "(nice_docnum): malloc failed.\n");
	exit(1);
    }

    /* copy alphanumeric chars of docnum into result */

    result_ptr = result;
    docnum_ptr = docnum;
    while((*docnum_ptr != '\0'))                /* at EOS ?                 */
    {
	    if (isdigit(*docnum_ptr))           /* should we copy ?         */
	    {
		    *result_ptr = *docnum_ptr;  /* copy!                    */
		    result_ptr++;               /* increment target pointer */
	    }
	    docnum_ptr++;                       /* increment source pointer */
    }
    *result_ptr = *docnum_ptr;                  /* stick on EOS char */
    return(result);
}

char *
case_strstr(char *s, char *c)
{
	char *result;
	char *sptr;
	char *sptr2;
	char *cptr;
	
	if( s == NULL ) return NULL;
	if( c == NULL ) return NULL;

	sptr = string_lower(s);
	cptr = string_lower(c);

	sptr2 = strstr(sptr, cptr);
	if( sptr2 )
		result = s + (sptr2 - sptr);
	else
		result = NULL;
		
	UFREE(sptr);
	UFREE(cptr);
	return result;
}

Bool
case_eq( char s, char c )
{
	return (muc_tolower(s) == muc_tolower(c));
}


/*******************************************************************
 *
 * muc_strcasecmp
 *
 *******************************************************************/

int
muc_strcasecmp(char *str1, char *str2)
{
	char *p1, *p2;

	if(str1==NULL && str2) return -1;
	if(str2==NULL) return 1;
	
	p1 = str1;
	p2 = str2;
	
	while(*p1 && *p2)
	{
		if( muc_tolower(*p1) != muc_tolower(*p2) )
			return muc_tolower(*p1) - muc_tolower(*p2);
		p1++;
		p2++;
	}
	return (*p1 - *p2);
}


char
muc_tolower(char c)
{
	if( c >= 'A' && c <= 'Z' )
		return (char) tolower((int)c);
	return c;
}
