/*
 *  Copyright 1995 Science Applications International Corporation.
 */

/*
 * NAME
 *
 * 
 * FILE 
 *	option.c
 *
 * SYNOPSIS
 *      option_create
 *      option_release
 *      option_compare
 *      option_add_value
 *      option_find
 *      option_find_values
 *      option_find_first_value
 *      option_table_add_defaults
 *      option_print_table
 *      option_print
 *
 * DESCRIPTION
 *	support routines for Option data structure.
 *
 * DIAGNOSTICS
 *
 *
 * FILES
 *
 *
 * NOTES
 *	Scratch pad - modification history, plans for enhancements.
 * 
 * SEE ALSO
 *
 * AUTHOR
 *	Gary Dungca	August 1995
 *
 */

#ifdef SCCSID
static char SccsId[] = "@(#)option.c	6.1\t09/08/97\tSAIC";
#endif

#undef SCCSID
#include  "muc_defs.h"
#include  "muc.h"
#include  "muc_protos.h"
#include  "muc_globals.h"

/************************************************************************
 *									*
 *  option_create
 *									*
 ***********************************************************************/

Option *
option_create(char *name)
{
    Option *option;
    
    if((option = UALLOC(Option, 1)) == (Option *)NULL)
    {
	fprintf(stderr, "Error: [option_create]: Allocation failed.\n");
	exit(1);
    }
    
    option->name = STRALLOC(name);

    /* NULL destroy function allows table_destroy() to empty table */
    option->values = table_list_create(cstring_compare,
				       default_release,
				       NULL);
    return option;
}

/************************************************************************
 *									*
 *  option_release
 *									*
 ***********************************************************************/

void
option_release(char *elt)
{
    Option *option = (Option *)elt;
    
    if(option == (Option *)NULL)
	return;

    UFREE(option->name);
    table_destroy(option->values);
    
    UFREE(option);
}


/************************************************************************
 *									*
 *  option_compare							*
 *									*
 ***********************************************************************/

int
option_compare(char *elt1, char *elt2)
{
    Option *option1 = (Option *)elt1;
    Option *option2 = (Option *)elt2;
    
    return(muc_strcasecmp(option1->name, option2->name));
}


/************************************************************************
 *									*
 *  option_find
 *									*
 *  given a table of options and a string, return the option named by
 *      the string, or NULL if there is none.
 *									*
 ***********************************************************************/

Option *
option_find(TABLE *option_table, char *optname)
{
    Option  *option;
    
    /* search the table for a option having the same name as optname */

    table_first(option_table);
    for(option = (Option *)table_next(option_table);
	option;
	option = (Option *)table_next(option_table))
      if( STRCASEEQ(optname, option->name) )
	break;

    return(option);
}


/************************************************************************
 *									*
 *  option_find_values
 *      Looks for the option named optname in the table, then returns 
 *      that option's table of values, or NULL if there's no option 
 *      by that name.
 *
 ***********************************************************************/

TABLE *
option_find_values(TABLE *option_table, char *optname)
{
    Option *option;

    option = option_find(option_table, optname);
    if( option == NULL )
      return NULL;
    
    return(option->values);
}

/************************************************************************
 *									*
 *  option_find_first_value
 *      returns the *first* value of the option named optname, or
 *      NULL if no values exist, or if no option is named optname
 *
 ***********************************************************************/

char *
option_find_first_value(TABLE *option_table, char *optname)
{
    TABLE *values;

    values = option_find_values(option_table, optname);
    if( values == NULL )
      return NULL;

    table_first( values );
    return( (char *)table_next(values) );
}


/************************************************************************
 *									*
 * option_print_table
 *									*
 ***********************************************************************/

void
option_print_table(TABLE *option_table, int indent)
{
  Option *option;
  table_first( option_table );
  for(option = (Option *)table_next(option_table);
      option;
      option = (Option *)table_next(option_table))
    option_print( option, indent );
}


/************************************************************************
 *									*
 * option_print
 *									*
 ***********************************************************************/

void
option_print(Option *option, int indent)
{
  char *value;

  printf("%*.s%s\n", indent, "", option->name );
  
  cstring_print_table( option->values, indent+1 );
}

