/*
 *  Copyright 1997 Science Applications International Corporation.
 */

/*
 * NAME
 * 
 * FILE 
 *	@(#)parse_templates.c	6.1
 *
 * SYNOPSIS
 *
 * DESCRIPTION
 *	template file (ST, TE, TR) parser
 *
 * DIAGNOSTICS
 *
 * FILES
 *
 * NOTES
 * 
 * SEE ALSO
 *
 * AUTHOR
 */

#ifdef SCCSID
static char SccsId[] = "@(#)parse_templates.c	6.1\t09/08/97\tSAIC";
#endif

#undef SCCSID
#include  "muc_defs.h"
#include  "muc.h"
#include  "muc_protos.h"
#include  "muc_globals.h"

#define COMMENT 1
#define BLANK   2
#define INST_START 3
#define SLOT_START 4
#define MFIL_START 5
#define NON_BLANK  6

static int  cur_lineno;
static char *cur_classname;
static char *cur_docnum;
static char *cur_instnum;
static char *cur_slotname;
static char *fname;
static Bool ignore_cur_inst;

static Catalog_entry *cur_catalog_entry;
static Clas *cur_clas;
static Inst *cur_inst;
static Slot *cur_slot;
static Mfil *cur_mfil;
static Sfil *cur_sfil;

static Class_def *cur_class_def;
static Slot_def *cur_slot_def;

/**************/

static void parse_cur_line( char *cur_line );
static char cur_line_type( char *cur_line );
static Bool is_inst_start( char *cur_line );
static Bool is_slot_start( char *cur_line );

static void parse_inst_start(char *bufptr );
static char *parse_pointer_stringrep(char *bufptr,
				     char **classname,
				     char **docnum,
				     char **instnum);
static void set_cur_class_def();

static void parse_slot_start(char *bufptr);
static char *parse_cur_slotname( char *bufptr );
static void set_cur_slot_def();

static void parse_mfil_start( char *bufptr );

static void parse_sfil_line( char *bufptr );
static void parse_stringfill_line( char *bufptr );
static void parse_setfill_line(char *bufptr);
static void parse_pointerfill_line( char *bufptr );
static void extract_fill_offsets(char *cur_line, 
				 int *start_offset, 
				 int *end_offset );
static Inst *find_pointed_to_inst(char *string_rep);

static void finish_cur_inst();
static void finish_cur_slot();
static void finish_cur_mfil();
static void finish_cur_sfil();
static void finish_cur_setfill();
static void finish_cur_stringfill();
static void finish_cur_pointerfill();

/*************************************************/
/*************************************************/

void
parse_templates(char whichfile)
{
	FILE *fopen();
	FILE *templatesin;
	char cur_line[MAX_MESSSIZE];
	
	cur_lineno = 1;
	MUC_key_or_response = whichfile;
	
	if ( MUC_key_or_response == KEY )
		fname = option_find_first_value(MUC_Options, ":key_file");
	else
		fname = option_find_first_value(MUC_Options, ":response_file");
	
	if (fname == NULL)
	{
		fprintf(stderr,
			"[parse_templates]: %s template file name not found.\n",
			( (MUC_key_or_response == KEY) ? "key" : "response"));
		exit(1);
	}
	
	if ((templatesin = fopen(fname, "r")) == (FILE *)NULL)
	{
		fprintf(stderr, "Could not open template file '%s'.\n",
			fname);
		exit(1);
	}

	ignore_cur_inst = False;
	
	while (!feof(templatesin))
	{
		if( fgets(cur_line, MAX_MESSSIZE, templatesin ) )
		{
			cur_lineno++;
		
			/* chomp newline */
			cur_line[strlen(cur_line) - 1] = '\0'; 

			parse_cur_line(cur_line);
		}
	}
	
	fclose(templatesin);
	
	return;
}

/*************************************************/
/*************************************************/

static
void 
parse_cur_line( char *cur_line )
{
	switch(cur_line_type(cur_line))
	{
	case COMMENT:
		break;
	case BLANK:
		break;
	case INST_START:
		ignore_cur_inst = False;
		parse_inst_start(cur_line);
		break;
	case SLOT_START:
		if( ! ignore_cur_inst )
			parse_slot_start(cur_line);
		break;
	case MFIL_START:
		if( ! ignore_cur_inst )
			parse_mfil_start(cur_line);
		break;
	case NON_BLANK:
		if( ! ignore_cur_inst )
			parse_sfil_line(cur_line);
		break;
	default:
		fprintf(stderr,
			"[parse_cur_line]: couldn't identify line %d\n",
			cur_lineno);
		exit(1);
	}
}

/*************************************************/
/*************************************************/

static
char
cur_line_type( char *cur_line )
{
	if( cur_line[0] == '#' || cur_line[0] == ';' )
		return COMMENT;

	/* strspn -- length of prefix consisting of chars in string */
	if( strspn( cur_line, " \t" ) == strlen(cur_line) )
		return BLANK;
	
	/* first non-blank char is slash */
	if( cur_line[strspn(cur_line, " \t")] == '/' )
		return MFIL_START;
	
	if( is_inst_start( cur_line ) )
		return INST_START;
	
	if( is_slot_start( cur_line ) )
		return SLOT_START;
	
	return NON_BLANK;
}

/*
 *
 * is_inst_start
 *     returns true if:
 *        o string's first non-blank char is '<'
 *        o there is a hyphen '-' following that
 *        o there is another hyphen following that
 *        o there is a '>' following that
 *        o after skipping any whitespace, there's a ":="
 *        o there is nothing else but whitespace after that
 *
 */

static
Bool
is_inst_start( char *cur_line )
{
	char *bufptr = cur_line;
	
	bufptr += strspn(bufptr, " \t");   /* skip whitespace */
	if( *bufptr != '<' ) return False; /* check open angle */
	
	bufptr += strcspn(bufptr, "-");    /* find first hyphen */
	if( *bufptr != '-' ) return False; /* check that it's found */	
	bufptr++;                          /* advance one char */

	bufptr += strcspn(bufptr, "-");    /* find second hyphen */
	if( *bufptr != '-' ) return False; /* check that it's found */	
	bufptr++;                          /* advance one char */

	bufptr += strcspn(bufptr, ">");    /* find close angle */
	if( *bufptr != '>' ) return False; /* check that it's found */	
	bufptr++;                          /* advance one char */

	bufptr += strspn( bufptr, " \t" ); /* skip whitespace */
	if((*bufptr == ':')                /* check for ":=" */
	   && 
	   (*(bufptr + 1) == '=') )
		bufptr += 2;
	else	
		return False;
	
	bufptr += strspn( bufptr, " \t" ); /* skip whitespace */
	if( *bufptr == '\0' )              /* nothing else on line */
		return True;
	else
		return False;
}


/*************************************************/
/*************************************************/

static
Bool
is_slot_start( char *cur_line )
{
	Bool result;
	char *bufptr=cur_line;   /* where in cur_line */
	char *slotname;          /* copy of slotname from cur_line */
	char *namestart;         /* pointer to start of slotname in cur_line */

	bufptr += strspn(bufptr, " \t");              /* skip whitespace */
	namestart = bufptr;                           /* save slotname start */
	
	if( !isalpha( *bufptr ) ) return False;       /* slotname first char */
	while( isalnum(*bufptr) || (*bufptr == '_') ) /* find slotname end   */
		bufptr++;
	slotname = alloc_between(namestart, bufptr);  /* copy slotname       */
	
	bufptr += strspn(bufptr, " \t");              /* skip whitespace */
	if( *bufptr != ':' )                          /* colon after name */
		result = False;
	else if( cur_class_def == NULL )              /* see if slot defined */
		result = False;
	else if( cur_class_def->slot_defs == NULL )
	{
		fprintf(stderr,
			"[is_slot_start]:null slot_defs in class_def\n");
		exit(1);
	}
	else if( slot_def_find(cur_class_def->slot_defs, 
			       slotname) )
		result = True;
	else
		result = False;
	UFREE(slotname);
	return result;
}

/*************************************************/
/*************************************************/

static
void
parse_inst_start(char *bufptr)
{
	finish_cur_inst();          /* Deallocates cur_classname,
   				                   cur_docnum,
						   cur_instnum.
				       Resets cur_class_def,	
				              cur_slot_def         */
	
	bufptr = parse_pointer_stringrep(bufptr, 
					 &cur_classname,	
					 &cur_docnum,
					 &cur_instnum );
	 
	if(cur_catalog_entry && 
	   STRCASEEQ(cur_catalog_entry->docnum, cur_docnum ))
		;  /* already have right catalog_entry */
	else
	{
		printf("%s ", cur_docnum);fflush(stdout);
		cur_catalog_entry = 
			catalog_entry_find( MUC_Catalog, cur_docnum );
		if( cur_catalog_entry == NULL )
		{
			cur_catalog_entry = catalog_entry_create(cur_docnum);
			table_append((char *)cur_catalog_entry,
				     MUC_Catalog);
		}
	}
	
	set_cur_class_def();

	if( MUC_key_or_response == KEY )
		cur_clas = 
			clas_find(cur_catalog_entry->key_classes,
				  cur_class_def->name);
	else	
		cur_clas = 
			clas_find(cur_catalog_entry->response_classes,
				  cur_class_def->name);
	if(cur_clas == NULL)
	{
		fprintf(stderr,
			"[parse_inst_start]: class %s not in document %s\n",
			cur_classname,
			cur_docnum);
		exit(1);
	}
	
	/* instance could be in catalog already, if something   */
	/* else points to it.  But it shouldn't be instantiated */
	/* yet.                                                 */
	cur_inst = inst_find(cur_clas->insts, cur_instnum);
	if( cur_inst && (cur_inst->is_instantiated) )
	{
		fprintf(stderr,
			"\nWARNING: ignoring repeat of %s object <%s-%s-%s>\n",
			(MUC_key_or_response==KEY)? "key" : "response",
			((cur_inst->name) ? (cur_inst->name) : "" ),
			((cur_inst->docnum) ? (cur_inst->docnum) : "" ),
			((cur_inst->id) ? (cur_inst->id) : "" ));
		ignore_cur_inst = True;
	}
	/* if instance wasn't in catalog at all, make it and put it in */
	else if( cur_inst == NULL )
	{
		cur_inst = inst_create(cur_classname,
				       cur_docnum,
				       cur_instnum);
		cur_inst->its_class_def = cur_class_def;
		if(MUC_key_or_response==KEY)
			cur_inst->is_key = True;
		table_append( (char *) cur_inst, cur_clas->insts );
	}
	cur_inst->is_instantiated = True;
}

/*************************************************/
/*************************************************/

static
char *
parse_pointer_stringrep(char *bufptr,
			 char **classname,
			 char **docnum,
			 char **instnum)
{
	char *tokenstart;
	bufptr += strspn(bufptr, " \t");   /* skip whitespace */
	if( *bufptr != '<' )
	{
		fprintf(stderr,
			"pointer missing '<' line %d\n",
			cur_lineno);
		exit(1);
	}
	bufptr++;                           /* go past '<'             */
	tokenstart = bufptr;                /* save start of classname */
	bufptr += strcspn( bufptr, "-" );   /* find first '-'          */
	if( *bufptr == '\0' )
	{
		fprintf(stderr,
			"pointer missing first '-', line %d\n",
			cur_lineno);
		exit(1);
	}
	*classname = alloc_between(tokenstart, bufptr);

	bufptr++;                           /* go past first '-'    */
	tokenstart = bufptr;                /* save start of docnum */
	bufptr += strcspn( bufptr, "-" );   /* find second '-'      */
	if( *bufptr == '\0' )
	{
		fprintf(stderr,
			"pointer missing second '-', line %d\n",
			cur_lineno);
		exit(1);
	}
	*docnum = alloc_between(tokenstart, bufptr);

	bufptr++;                           /* go past second '-'    */
	tokenstart = bufptr;                /* save start of docnum  */
	bufptr += strcspn( bufptr, ">" );   /* find second '-'       */
	if( *bufptr == '\0' )
	{
		fprintf(stderr,
			"pointer missing '>', line %d\n",
			cur_lineno);
		exit(1);
	}
	*instnum = alloc_between(tokenstart, bufptr);

	bufptr++;                           /* go past '>' */
	return bufptr;
}

/*************************************************/
/*************************************************/

static
void
set_cur_class_def()
{
	if(cur_classname == NULL)
	{
		fprintf(stderr,
			"[set_cur_class_def]: cur_classname is NULL\n");
		exit(1);
	}
	cur_class_def = class_def_find(MUC_Class_defs, 
				       cur_classname);
	if(cur_class_def == NULL)
	{
		fprintf(stderr,
			"object type \'%s\' undefined, line %d\n",	
			cur_classname,
			cur_lineno);
		exit(1);
	}
		
}


/*************************************************/
/*************************************************/

static
void
parse_slot_start(char *bufptr)
{
	if(cur_inst == NULL)
	{
		fprintf(stderr,
			"[parse_slot_start]: No current object %s line %s\n",
			fname,
			cur_lineno);
		exit(1);
	}
	
	finish_cur_slot();
	
	bufptr = parse_cur_slotname(bufptr);	/* advances bufptr past ':' */
	set_cur_slot_def();

	cur_slot = slot_find( cur_inst->slots, cur_slotname );
	if( cur_slot )
	{
		fprintf(stderr,
			"Warning: redefining slot %s on line %d\n",
			cur_slotname,
			cur_lineno);
		table_delete( (char *) cur_slot, cur_inst->slots );
	}
	cur_slot = slot_create(cur_slotname);
	cur_slot->its_slot_def = cur_slot_def;
	table_append( (char *)cur_slot, cur_inst->slots );
	
	if( *bufptr == '\0' ) return;   /* finished line */

	bufptr += strspn(bufptr, " \t");           /* skip whitespace */
	if( *bufptr == '/' )
	{
		cur_slot->is_optional = True;
		bufptr++;                          /* skip '/'        */
		bufptr += strspn(bufptr, " \t");   /* skip whitespace */
		if( *bufptr == '\0' ) return;      /* reached EOL   */
	}
	parse_mfil_start( bufptr );
}

/*************************************************/
/*************************************************/

static
char *
parse_cur_slotname( char *bufptr )
{
	char *slotname_start;
		
	bufptr += strspn(bufptr, " \t");  /* skip whitespace */

	slotname_start = bufptr;          /* save start of name */
	
	if( !isalpha( *bufptr ) )            /* name should start w/ letter */
	{
		fprintf(stderr,
			"[parse_cur_slotname]: invalid slotname, line %d\n",
			cur_lineno);
		exit(1);
	}
	bufptr++;                          /* advance past first letter */

	while( isalnum(*bufptr) || (*bufptr == '_') )
		bufptr++;                  /* go past end of slotname */

	cur_slotname = alloc_between( slotname_start, bufptr );
	
	bufptr += strspn(bufptr, " \t");  /* skip whitespace */

	if( *bufptr != ':' )                 /* next char should be colon */
	{
		fprintf( stderr,
			"[parse_cur_slotname]: no colon after slotname, line %d\n",
			cur_lineno);
		exit(1);
	}
	bufptr++;                           /* advance past colon */
	return bufptr;
}

/*************************************************/
/*************************************************/

static
void
set_cur_slot_def()
{
	if( cur_slotname == NULL )
	{
		fprintf(stderr,
			"[set_cur_slot_def]: cur_slotname is NULL\n");
		exit(1);
	}
	if( cur_class_def == NULL )
	{
		fprintf(stderr,
			"[set_cur_slot_def]: cur_class_def is NULL\n");
		exit(1);
	}
	cur_slot_def = 
		slot_def_find(cur_class_def->slot_defs, 
			      cur_slotname);
}

	
/*************************************************/
/*************************************************/

static
void
parse_mfil_start( char *bufptr )
{
	if( cur_slot == NULL )
	{
		fprintf(stderr,
			"[parse_mfil_slot]: No current slot, line %d\n",
			cur_lineno);
		exit(1);
	}
	finish_cur_mfil();
	bufptr += strspn(bufptr, " \t/");   /* skip whitespace and '/'s */
	
	cur_mfil = mfil_create();
	table_append( (char *) cur_mfil, cur_slot->mfils );
	
	if( *bufptr == '\0' )               /* nothing left on line */
		return;
	else
		parse_sfil_line( bufptr );  /* still single fill(s) on line */
}

/*************************************************/
/*************************************************/

static
void
parse_sfil_line( char *bufptr )
{
	char *slot_type;
	
	if( cur_mfil == NULL || cur_slot == NULL )
	{
		fprintf(stderr,
			"Error: Data on line %d not in a slot.\n",
			cur_lineno - 1);
		exit(1);
	}
	finish_cur_sfil();
	
	table_first(cur_slot_def->fill_types);
	slot_type = table_next(cur_slot_def->fill_types);
	if(STRCASEEQ( slot_type, "string" ))
		parse_stringfill_line(bufptr);
	else if(STRCASEEQ( slot_type, "set" ))
		parse_setfill_line(bufptr);
	else
		parse_pointerfill_line(bufptr);
}

/*************************************************/
/*************************************************/

static
void
parse_stringfill_line( char *bufptr )
{
	int start_offset;
	int end_offset;
	
	extract_fill_offsets( bufptr, &start_offset, &end_offset );
	cur_sfil = sfil_create(STRING_FILL);
	cur_sfil->fill.str->fill = trim_quotes( bufptr );
	cur_sfil->fill.str->start_offset = start_offset;
	cur_sfil->fill.str->end_offset = end_offset;
	table_append( (char *) cur_sfil, cur_mfil->sfils );
}

/*************************************************/
/*************************************************/

static
void
parse_setfill_line(char *bufptr)
{
	int start_offset;
	int end_offset;
	
	extract_fill_offsets( bufptr, &start_offset, &end_offset );
	
	cur_sfil = sfil_create(SET_FILL);
	cur_sfil->fill.set->fill = trim_quotes( bufptr );
	cur_sfil->fill.set->start_offset = start_offset;
	cur_sfil->fill.set->end_offset = end_offset;
	table_append( (char *) cur_sfil, cur_mfil->sfils );
}

/*************************************************/
/*************************************************/

static
void
parse_pointerfill_line( char *bufptr )
{
	char *string_rep;       /* string representation of pointer */
	char *tokenstart;
	
	int start_offset;
	int end_offset;
	
	extract_fill_offsets( bufptr, &start_offset, &end_offset );
	
	bufptr += strspn(bufptr, " \"\'\t");  /* skip ws and open quotes */
	
	while( *bufptr )
	{
		if( *bufptr != '<' )
		{
			fprintf(stderr,
				"syntax error, line %d: expected pointer\n",
				cur_lineno);
			exit(1);
		}
		tokenstart = bufptr;                /* save '<' position */
		bufptr++;                           /* go past '<'       */
		bufptr += strcspn( bufptr, ">" );   /* find '>'          */

		if( *bufptr != '>' )
		{
			fprintf(stderr,
				"syntax error, line %d: no pointer close\n",
				cur_lineno);
			exit(1);
		}
		bufptr++;                           /* go past '>'      */

		string_rep = alloc_between( tokenstart, bufptr );
		cur_sfil = sfil_create(POINTER_FILL);
		cur_sfil->fill.ptr->string_rep = string_rep;
		cur_sfil->fill.ptr->fill = find_pointed_to_inst(string_rep);
		cur_sfil->fill.ptr->start_offset = start_offset;
		cur_sfil->fill.ptr->end_offset = end_offset;
		table_append( (char *)cur_sfil, cur_mfil->sfils );

		bufptr += strspn(bufptr, " \"\'\t");  /* skip ws, quotes */
	}
}

/*************************************************/
/*************************************************/

static
void
extract_fill_offsets(char *cur_line, int *start_offset, int *end_offset )
{
#define BAILOUT {UFREE(int_string);(*start_offset)=(*end_offset)=0;return;}
	 
	char *non_integer = NULL;
	char *int_string = NULL;
	char *start_first, *start_last, *end_first, *end_last;
	char *bufptr;
	
	bufptr = cur_line + strlen(cur_line) - 1;   /* last char */

	
	while(((*bufptr) != '#')                  /* search backward for '#' */
	      && 
	      (bufptr >= cur_line))
		bufptr--;
	if( bufptr <= cur_line ) BAILOUT

	end_last = bufptr;                     /* save char after end offset */
	bufptr--;                              /* backup to last offset char */
	
	while(((*bufptr) != '#')                  /* search backward for '#' */
	      && 
	      (bufptr >= cur_line))
		bufptr--;
	if( bufptr <= cur_line ) BAILOUT

	end_first = bufptr + 1;             /* save first char of end offset */
	start_last = bufptr;                /* save char after start offset  */
	bufptr--;                           /* back up to last offset char   */
	
	while(((*bufptr) != '#')                  /* search backward for '#' */
	      && 
	      (bufptr >= cur_line))
		bufptr--;
	if( bufptr <= cur_line ) BAILOUT
	
	start_first = bufptr + 1;         /* save first char of start offset */
	bufptr--;
	if( *bufptr != '#' ) BAILOUT     /* should be first of double-'#'   */

	int_string = alloc_between(end_first, end_last);
	*end_offset = strtol( int_string, &non_integer, 10 );
	if( non_integer && (*non_integer) ) BAILOUT /* string had non-digits */
	UFREE(int_string);

	int_string = alloc_between(start_first, start_last);
	*start_offset = strtol( int_string, &non_integer, 10 );
	if( non_integer && (*non_integer) ) BAILOUT
	UFREE(int_string);

	*bufptr = '\0';                 /* truncate line at offset info */
}

/*************************************************/
/*************************************************/

static
Inst *
find_pointed_to_inst(char *string_rep)
{
	char *docnum;
	char *classname;
	char *instnum;

	Catalog_entry *catalog_entry;
	TABLE *classes;
	Class_def *class_def;
	Clas *clas;
	Inst *inst;
	
	parse_pointer_stringrep(string_rep,
				&classname,
				&docnum,
				&instnum);

	if( cur_catalog_entry == NULL )
	{
		fprintf(stderr,
			"No current catalog entry\n");
		exit(1);
	}

	if(STRCASEEQ(cur_catalog_entry->docnum, docnum))
		catalog_entry = cur_catalog_entry;
	else 	
	{
		fprintf(stderr,
			"Pointer\'s document number is wrong: ");
		fprintf(stderr, 
			"%s, line %d\n", 
			string_rep,
			cur_lineno);
		exit(1);
	}
	
	if( MUC_key_or_response == KEY )
		classes = catalog_entry->key_classes;
	else
		classes = catalog_entry->response_classes;
	
	clas = clas_find(classes, classname);
	if( clas == NULL )
	{
		fprintf(stderr,
			"Class %s (line %d) not found in catalog\n",
			classname,
			cur_lineno);
		exit(1);
	}
	
	inst = inst_find(clas->insts, instnum);
	if( inst == NULL )
	{
		inst = inst_create(classname, docnum, instnum);
		inst->is_pointed_to = True;
		inst->its_class_def =
			class_def_find(MUC_Class_defs,
				       classname);
		if(MUC_key_or_response==KEY)
			inst->is_key = True;
		table_append((char *)inst, clas->insts);
	}
	UFREE(classname);
	UFREE(docnum);
	UFREE(instnum);
	return inst;
}

/*************************************************/
/*************************************************/


static
void
finish_cur_inst()
{
	if( cur_inst == NULL ) return;
	
	finish_cur_slot();
	
	if(table_count(cur_inst->slots) == 0)
		;    /* nothing -- it's ok to have empty insts!? */
	
	cur_inst = NULL;
	UFREE( cur_classname );
	UFREE( cur_docnum );
	UFREE( cur_instnum );
	cur_class_def = NULL;
}

/*************************************************/
/*************************************************/

static
void
finish_cur_slot()
{
	if( cur_slot == NULL ) return;
	
	finish_cur_mfil();
	
	if( table_count( cur_slot->mfils ) == 0 )
		table_delete((char *)cur_slot, cur_inst->slots);

	UFREE(cur_slotname);
	cur_slot_def = NULL;
	cur_slot = NULL;
}

/*************************************************/
/*************************************************/

static
void
finish_cur_mfil()
{
	if(cur_mfil == NULL) return;
	
	finish_cur_sfil();
	
	if(table_count(cur_mfil->sfils) == 0)
		table_delete((char *) cur_mfil, cur_slot->mfils);
	
	cur_mfil = NULL;
}

/*************************************************/
/*************************************************/

static
void
finish_cur_sfil()
{
	if(cur_sfil == NULL ) return;
	
	switch( cur_sfil->type )
	{
	case SET_FILL:
		finish_cur_setfill();
		break;
	case POINTER_FILL:
		finish_cur_pointerfill();
		break;
	case STRING_FILL:
		finish_cur_stringfill();
		break;
	default:
		fprintf(stderr,
			"[finish_cur_sfil]: sfil type %d unknown, line %d\n",
			cur_sfil->type,
			cur_lineno);
		exit(1);
	}
	cur_sfil = NULL;
}

/*************************************************/
/*************************************************/

static
void
finish_cur_setfill()
{
	return;
}

/*************************************************/
/*************************************************/

static
void
finish_cur_stringfill()
{
	return;
}
/*************************************************/
/*************************************************/

static
void
finish_cur_pointerfill()
{
	return;
}


