/*
 *  Copyright 1995 Science Applications International Corporation.
 */

/*
 * NAME
 *
 * 
 * FILE 
 *	partitionset.c
 *
 * SYNOPSIS
 *
 *
 * DESCRIPTION
 *	support routines for Partitionset data structure.
 *
 * DIAGNOSTICS
 *
 *
 * FILES
 *
 *
 * NOTES
 *	Scratch pad - modification history, plans for enhancements.
 * 
 * SEE ALSO
 *
 * AUTHOR
 *	Gary Dungca	August 1995
 *
 */


#ifdef SCCSID
static char SccsId[] = "@(#)partitionset.c	6.1\t09/08/97\tSAIC";
#endif

#undef SCCSID
#include  "muc_defs.h"
#include  "muc.h"
#include  "muc_protos.h"
#include  "muc_globals.h"

static void partitionset_init_array(TABLE **tables,
				    int arraysize);
static void partitionset_compute_partitions(Partitionset *partitionset,
					    Eqclasset *key_eqclasset,
					    Eqclasset *resp_eqclasset,
					    TABLE *inst_pairs);




/************************************************************************
 *									*
 *  partitionset_create
 *									*
 ***********************************************************************/

Partitionset *
partitionset_create(Eqclasset *key_eqclasset, 
		    Eqclasset *resp_eqclasset, 
		    TABLE *inst_pairs)
{
	Partitionset *partitionset;

	if(key_eqclasset == NULL
	   || 
	   resp_eqclasset == NULL
	   || 
	   inst_pairs == NULL)
		return NULL;
	
  
	if((partitionset = UALLOC(Partitionset, 1)) == (Partitionset *)NULL)
	{
		fprintf(stderr, 
			"Error: [partitionset_create]: Allocation failed.\n");
		exit(1);
	}
  
	partitionset->num_key_classes = 
		table_count(key_eqclasset->eqclasses);
	partitionset->num_resp_classes = 
		table_count(resp_eqclasset->eqclasses);


	partitionset->matched = 
		(TABLE **)calloc((partitionset->num_key_classes)
				 *
				 (partitionset->num_resp_classes),
				 sizeof(TABLE **));
	partitionset->missing = 
		(TABLE **)calloc((partitionset->num_key_classes),
				 sizeof(TABLE **));
	partitionset->spurious = 
		(TABLE **)calloc((partitionset->num_resp_classes),
				 sizeof(TABLE **));


	partitionset_init_array(partitionset->matched,
				(partitionset->num_key_classes)
				*
				(partitionset->num_resp_classes));
	partitionset_init_array(partitionset->missing,
				(partitionset->num_key_classes));
	partitionset_init_array(partitionset->spurious,
				(partitionset->num_resp_classes));
	
	partitionset_compute_partitions(partitionset, 
					key_eqclasset,
					resp_eqclasset,
					inst_pairs);

	return(partitionset);
}

/************************************************************************
 *									*
 *  partitionset_init_array
 *									*
 ***********************************************************************/

static
void
partitionset_init_array(TABLE **tables,
			int arraysize)
{
	int i;
	
	for(i=0; i<arraysize; i++)
		tables[i] = table_list_create(default_compar,
					      NULL,
					      default_destroy);
}

/************************************************************************
 *									*
 *  partitionset_compute_partitions
 *									*
 ***********************************************************************/

static
void
partitionset_compute_partitions(Partitionset *partitionset,
				Eqclasset *key_eqclasset,
				Eqclasset *resp_eqclasset,
				TABLE *inst_pairs)
{
	Inst_pair *inst_pair;
	Inst *key_inst;
	Inst *resp_inst;
	Eqclass *key_eqclass;
	Eqclass *resp_eqclass;
	int key_index;
	int resp_index;
	int matched_index;

	eqclass_compute_indices(key_eqclasset->eqclasses);
	eqclass_compute_indices(resp_eqclasset->eqclasses);

	table_first(inst_pairs);
	for(inst_pair=(Inst_pair *)table_next(inst_pairs);
	    inst_pair;
	    inst_pair=(Inst_pair *)table_next(inst_pairs))
	{
		key_inst = inst_pair->key_inst;
		resp_inst = inst_pair->response_inst;
		if(key_inst)
		{
			key_eqclass = lookup_eqclass(key_inst, 
						     key_eqclasset);
			if( key_eqclass )
				key_index = key_eqclass->class_num;
			else
				continue;  /* no warning, since this */
			                   /* oft happens with OPT's */
		}
		if(resp_inst)
		{
			resp_eqclass = lookup_eqclass(resp_inst,
						      resp_eqclasset);
			if( resp_eqclass )
				resp_index = resp_eqclass->class_num;
			else
				continue;
		}
		if( key_inst && resp_inst )
		{
			matched_index = 
				(key_index * partitionset->num_resp_classes)
					+ resp_index;
			table_append((char *)inst_pair,
				     partitionset->matched[matched_index]);
		}
		else if( key_inst )
			table_append((char *)inst_pair,
				     partitionset->missing[key_index]);
		else if( resp_inst )
			table_append((char *)inst_pair,
				     partitionset->spurious[resp_index]);
		else
			fprintf(stderr,
				"Warning: inst_pair w/out insts!\n");
	}
}

/************************************************************************
 *									*
 * partitionset_score_recall
 *									*
 ***********************************************************************/

void
partitionset_score_recall(Partitionset *ps,
			  long *numerator,
			  long *denominator)
{
	int key_index;
	int resp_index;
	int matched_index;
	int num_partitions;
	int partitionsize;
	int num_missing;
	
	(*numerator) = (*denominator) = 0;
	
	for(key_index=0;
	    key_index < ps->num_key_classes;
	    key_index++)
	{
		num_partitions = 0;
		for(resp_index=0;
		    resp_index < ps->num_resp_classes;
		    resp_index++)
		{
			matched_index = 
				(key_index * ps->num_resp_classes) 
					+ resp_index;
			partitionsize = 
				table_count(ps->matched[matched_index]);
			
			if( partitionsize > 0 )
			{
				(*denominator) += partitionsize;
				num_partitions++;
			}
		}
		num_missing = table_count(ps->missing[key_index]);
		num_partitions += num_missing;
		(*denominator) += num_missing;
		(*numerator)   += (num_partitions - 1);
	}
	(*numerator) = (*denominator) - (*numerator);
	(*numerator) -= (ps->num_key_classes);
	(*denominator) -= (ps->num_key_classes);
}
		       
/************************************************************************
 *									*
 * partitionset_score_precision
 *									*
 ***********************************************************************/

void
partitionset_score_precision(Partitionset *ps,
			     long *numerator,
			     long *denominator)
{
	int key_index;
	int resp_index;
	int matched_index;
	int num_partitions;
	int partitionsize;
	int num_spurious;
	
	(*numerator) = (*denominator) = 0;
	
	for(resp_index=0;
	    resp_index < ps->num_resp_classes;
	    resp_index++)
	{
		num_partitions = 0;
		for(key_index=0;
		    key_index < ps->num_key_classes;
		    key_index++)
		{
			matched_index = 
				(key_index * ps->num_resp_classes) 
					+ resp_index;
			partitionsize = 
				table_count(ps->matched[matched_index]);
			
			if( partitionsize > 0 )
			{
				(*denominator) += partitionsize;
				num_partitions++;
			}
		}
		num_spurious = table_count(ps->spurious[resp_index]);
		num_partitions += num_spurious;
		(*denominator) += num_spurious;
		(*numerator)   += (num_partitions - 1);
	}
	(*numerator) = (*denominator) - (*numerator);
	(*numerator) -= (ps->num_resp_classes);
	(*denominator) -= (ps->num_resp_classes);
}
		       

/************************************************************************
 *									*
 * partitionset_release
 *									*
 ***********************************************************************/

void
partitionset_release(char *c_ptr)
{
    Partitionset *partitionset = (Partitionset *)c_ptr;
    int i;
    
    if(partitionset == (Partitionset *)NULL)
	return;
    
    for(i=0; i<partitionset->num_key_classes; i++)
	    table_destroy(partitionset->missing[i]);
    for(i=0; i<partitionset->num_resp_classes; i++)
	    table_destroy(partitionset->spurious[i]);
    for(i=0; 
	i < (partitionset->num_key_classes * partitionset->num_resp_classes); 
	i++)
	    table_destroy(partitionset->matched[i]);

    
    UFREE(partitionset->missing);
    UFREE(partitionset->spurious);
    UFREE(partitionset->matched);
    UFREE(partitionset);
}


/************************************************************************
 *									*
 *  partitionset_compare					      
 *									*
 ***********************************************************************/

int
partitionset_compare(char *elt1, char *elt2)
{
    Partitionset *partitionset1 = (Partitionset *)elt1;
    Partitionset *partitionset2 = (Partitionset *)elt2;
    
    return((long)partitionset1 - (long)partitionset2);
}

